## Interface: {interface}
## Title: WeakAuras Companion
## Author: The WeakAuras Team
## Version: {version}
## Notes: Keep your WeakAuras updated!
## X-Category: Interface Enhancements
## X-Generator: instawow
## DefaultState: Enabled
## LoadOnDemand: 0
## OptionalDeps: WeakAuras, Plater

data.lua
init.lua
